import os
import sys
import pickle
import copy
import numpy as np
from scipy.spatial.transform import Rotation


def main():
    from TB2J.io_merge import merge
    merge(sys.argv[1], sys.argv[2], sys.argv[3], save=True)


main()



import numpy as np
from ase.atoms import Atoms
from TB2J.utils import symbol_number
from collections import defaultdict
from scipy.linalg import eigh

class SislWrapper():
    def __init__(self, sisl_hamiltonian, spin=None):
        self.ham = sisl_hamiltonian
        # k2Rfactor : H(k) = \int_R H(R) * e^(k2Rfactor * k.R)
        self.R2kfactor = -2.0j*np.pi   # 
        if spin=='up':
            spin=0
        elif spin=='down':
            spin=1
        if spin not in [ None, 0, 1, 'merge']:
            raise ValueError("spin should be None/0/1, but is %s"%spin)
        self.spin=spin
        self.orbs=[]
        self.orb_dict=defaultdict(lambda:[])
        g=self.ham._geometry
        _atoms=self.ham._geometry._atoms
        atomic_numbers=[]
        self.positions=g.xyz
        self.cell=np.array(g.sc.cell)
        for ia, a in enumerate(_atoms):
            atomic_numbers.append(a.Z)
        self.atoms=Atoms(numbers=atomic_numbers, cell=self.cell, positions=self.positions)
        sdict=list(symbol_number(self.atoms).keys())
        if self.ham.spin.is_colinear and (self.spin in [0,1]):
            for ia, a in enumerate(_atoms):
                symnum=sdict[ia]
                try:
                    orb_names=[f"{symnum}|{x.name()}|up" for x in a.orbital]
                except:
                    orb_names=[f"{symnum}|{x.name()}|up" for x in a.orbitals]
                self.orbs+=orb_names
                self.orb_dict[ia]+=orb_names
            self.norb = len(self.orbs)
            self.nbasis=self.norb
        elif self.ham.spin.is_spinorbit or self.spin=='merge':
            for spin in {'up', 'down'}:
                for ia, a in enumerate(_atoms):
                    symnum=sdict[ia]
                    orb_names=[]
                    try:
                        for x in a.orbital:  # sisl < v0.10
                            orb_names.append(f"{symnum}|{x.name()}|{spin}")
                    except:
                        for x in a.orbitals:  # sisl >= v0.10
                            orb_names.append(f"{symnum}|{x.name()}|{spin}")
                    self.orbs+=orb_names
                    self.orb_dict[ia]+=orb_names
            #print(self.orb_dict)
            self.norb=len(self.orbs)//2
            #print(f"Norb: {self.norb}")
            self.nbasis= len(self.orbs)
        else:
            raise ValueError("The hamiltonian should be either spin-orbit or colinear")

    def view_info(self):
        print(self.orb_dict)
        print(self.atoms)

    def solve(self, k, convention=2):
        if convention==1:
            gauge='r'
        elif convention==2:
            gauge='R'
        if self.spin in [0,1]:
            evals, evecs = self.ham.eigh(k=k, spin=self.spin, eigvals_only=False, gauge=gauge)
        elif self.spin is None:
            evals, evecs = self.ham.eigh(k=k, eigvals_only=False, gauge=gauge)
            # reorder the first (basis) dimension so that it is 1up,1down, 2up, 2down...
            evecs=np.vstack([evecs[::2, :], evecs[1::2,:]])
        elif self.spin=='merge':
            evals0, evecs0 = self.ham.eigh(k=k, spin=0, eigvals_only=False, gauge=gauge)
            evals1, evecs1 = self.ham.eigh(k=k, spin=1, eigvals_only=False, gauge=gauge)
            evals=np.zeros(self.nbasis, dtype=float)
            evecs=np.zeros((self.nbasis, self.nbasis), dtype=complex)
            evals[:self.norb]=evals0
            evals[self.norb:]=evals1
            evecs[:self.norb, :self.norb]=evecs0
            evecs[self.norb:, self.norb:]=evecs1
        return evals, evecs


    def Hk(self, k, convention=2):
        if convention==1:
            gauge='r'
        elif convention==2:
            gauge='R'
        if self.spin is None:
            H=self.ham.Hk(k, gauge=gauge,format='dense')
            H=np.vstack([H[::2,:], H[1::2,:]])
            H=np.hstack([H[:,::2], H[:,1::2]])
        elif self.spin in [0,1]:
            H=self.ham.Hk(k, spin=self.spin, gauge=gauge, format='dense')
        elif self.spin == 'merge':
            H=np.zeros((self.nbasis, self.nbasis), dtype='complex')
            H[:self.norb, :self.norb]=self.ham.Hk(k, spin=0, gauge=gauge, format='dense')
            H[self.norb:, self.norb:]=self.ham.Hk(k, spin=1, gauge=gauge, format='dense')
        return H

    def eigen(self, k, convention=2):
        return self.solve(k)

    def gen_ham(self, k, convention=2):
        return self.Hk(k, convention=convention)

    def Sk(self, k, convention=2):
        if convention==1:
            gauge='r'
        elif convention==2:
            gauge='R'
        S0=self.ham.Sk(k, gauge='R', format='dense')
        #print(f"shape:{S0.shape}")
        #print(f"{self.nbasis}")
        if self.spin is None:
            S=np.vstack([S0[::2,:], S0[1::2,:]])
            S=np.hstack([S[:,::2], S[:,1::2]])
            #S=np.zeros((self.nbasis, self.nbasis), dtype='complex')
            #S[:self.norb,:self.norb]=S0
            #S[self.norb:, self.norb:]=S0
            #S=np.zeros((self.nbasis, self.nbasis), dtype='complex')
            #S[:self.nbasis//2,:self.norb//2]=S0
            #S[self.norb//2:, self.norb//2:]=S0
        elif self.spin in [0,1]:
            S=S0
        elif self.spin=='merge':
            S=np.zeros((self.nbasis, self.nbasis), dtype='complex')
            S[:self.norb,:self.norb]=S0
            S[self.norb:, self.norb:]=S0
        return S

    def solve_all(self, kpts, orth=True):
        evals = []
        evecs = []
        for ik, k in enumerate(kpts):
            if orth:
                S=self.Sk(k)
                Smh=Lowdin(S)
                H=self.gen_ham(k)
                Horth=Smh.T.conj() @ H @ Smh
                evalue, evec = eigh(Horth)
            else:
                evalue, evec = self.solve(k)
            evals.append(evalue)
            evecs.append(evec)
        return np.array(evals, dtype=float), np.array(evecs, dtype=complex, order='C')

    def HS_and_eigen(self, kpts, convention=2):
        evals = []
        evecs = []
        H=[]
        S=[]
        for ik, k in enumerate(kpts):
            Hk = self.Hk(k, convention=convention)
            Sk = self.Sk(k, convention=convention)
            H.append(self.Hk(k, convention=convention))
            S.append(self.Sk(k, convention=convention))
            evalue, evec = self.solve(k, convention=convention)
            evals.append(evalue)
            evecs.append(evec)
        return np.array(H), np.array(S), np.array(evals, dtype=float), np.array(evecs, dtype=complex, order='C')


    def get_fermi_level(self):
        return 0.0



#def test():
#    fdf = sisl.get_sile('/home/hexu/projects/learn_siesta/SMO_Wannier/siesta.fdf')
#    H = fdf.read_hamiltonian(order='nc',dim=2)
#    print(H._spin._is_polarized)
#    print(H.__dict__.keys())
#    print(H._geometry.__dict__.keys())
#    print(H._geometry.xyz)
#    print(H._geometry.sc)
#    H._geometry.sc.cell
#    orb=H._geometry._atoms[0].orbital[0]
#    print(orb.name())
#    s=SislWrapper(H)
#    s.view_info()
#
#import sisl
#fdf = sisl.get_sile('/home/hexu/projects/learn_siesta/SMO_Wannier/siesta.fdf')
#H = fdf.read_hamiltonian(order='nc',dim=2)
#if __name__=='__main__':
#    test()



import os
from TB2J.myTB import MyTB, merge_tbmodels_spin
import numpy as np
from TB2J.exchange import ExchangeCL, ExchangeNCL
from TB2J.exchangeCL2 import ExchangeCL2
from TB2J.utils import read_basis, auto_assign_basis_name
from ase.io import read
from TB2J.sisl_wrapper import SislWrapper
from TB2J.gpaw_wrapper import GPAWWrapper

def gen_exchange(path,
                 colinear=True,
                 orb_order=1,
                 posfile='POSCAR',
                 prefix_up='wannier90.up',
                 prefix_dn='wannier90.dn',
                 prefix_SOC='wannier90',
                 min_hopping_norm=1e-4,
                 max_distance=None,
                 efermi=0,
                 magnetic_elements=[],
                 kmesh=[4, 4, 4],
                 emin=-12.0,
                 emax=0.0,
                 nz=100,
                 height=0.2,
                 nz1=50,
                 nz2=200,
                 nz3=50,
                 exclude_orbs=[],
                 Rcut=None,
                 ne=None,
                 description=''):
    atoms = read(os.path.join(path, posfile))
    basis_fname = os.path.join(path, 'basis.txt')
    if colinear:
        print("Reading Wannier90 hamiltonian: spin up.")
        tbmodel_up = MyTB.read_from_wannier_dir(
            path=path, prefix=prefix_up, posfile=posfile, nls=False)
        print("Reading Wannier90 hamiltonian: spin down.")
        tbmodel_dn = MyTB.read_from_wannier_dir(
            path=path, prefix=prefix_dn, posfile=posfile, nls=False)
        if os.path.exists(basis_fname):
            basis = read_basis(basis_fname)
        else:
            basis, _ = auto_assign_basis_name(tbmodel_up.xred, atoms)
        print("Starting to calculate exchange.")
        exchange = ExchangeCL2(
            tbmodels=(tbmodel_up, tbmodel_dn),
            atoms=atoms,
            basis=basis,
            efermi=efermi,
            magnetic_elements=magnetic_elements,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            nz=nz,
            height=height,
            nz1=nz1,
            nz2=nz2,
            nz3=nz3,
            exclude_orbs=exclude_orbs,
            Rcut=Rcut,
            ne=ne,
            description=description)
        exchange.run()
        print("All calculation finsihed. The results are in TB2J_results directory.")

    elif colinear and 0:
        print("Reading Wannier90 hamiltonian: spin up.")
        tbmodel_up = MyTB.read_from_wannier_dir(
            path=path, prefix=prefix_up, posfile=posfile, nls=False)
        print("Reading Wannier90 hamiltonian: spin down.")
        tbmodel_dn = MyTB.read_from_wannier_dir(
            path=path, prefix=prefix_dn, posfile=posfile, nls=False)
        tbmodel = merge_tbmodels_spin(tbmodel_up, tbmodel_dn)
        if os.path.exists(basis_fname):
            basis = read_basis(basis_fname)
        else:
            basis, _ = auto_assign_basis_name(tbmodel.xred, atoms)
        print("Starting to calculate exchange.")
        exchange = ExchangeCL(
            tbmodels=tbmodel,
            atoms=atoms,
            basis=basis,
            efermi=efermi,
            magnetic_elements=magnetic_elements,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            nz=nz,
            height=height,
            nz1=nz1,
            nz2=nz2,
            nz3=nz3,
            exclude_orbs=exclude_orbs,
            Rcut=Rcut,
            ne=ne,
            description=description)
        exchange.run()
        print("All calculation finsihed. The results are in TB2J_results directory.")
    else:
        print("Reading Wannier90 hamiltonian: non-colinear spin.")
        tbmodel = MyTB.read_from_wannier_dir(
            path=path, prefix=prefix_SOC, posfile=posfile, nls=True)
        if orb_order==1:
            pass
        if orb_order==2:
            tbmodel=tbmodel.reorder()
        if os.path.exists(basis_fname):
            print("The use of basis file is deprecated. It will be ignored.")
            #basis = read_basis(basis_fname)
        else:
            basis, _ = auto_assign_basis_name(tbmodel.xred, atoms)
        print("Starting to calculate exchange.")
        exchange = ExchangeNCL(
            tbmodels=tbmodel,
            atoms=atoms,
            basis=basis,
            efermi=efermi,
            magnetic_elements=magnetic_elements,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            nz=nz,
            height=height,
            nz1=nz1,
            nz2=nz2,
            nz3=nz3,
            exclude_orbs=exclude_orbs,
            Rcut=Rcut,
            ne=ne,
            description=description)
        print("\n")
        exchange.run()
        print("All calculation finsihed. The results are in TB2J_results directory.")


def gen_exchange_siesta(
                        fdf_fname,
                        magnetic_elements=[],
                        kmesh=[4, 4, 4],
                        emin=-12.0,
                        emax=0.0,
                        nz=50,
                        #height=0.2,
                        #nz1=50,
                        #nz2=200,
                        #nz3=50,
                        exclude_orbs=[],
                        Rcut=None,
                        ne=None,
                        description=''):

    try:
        import sisl
    except:
        raise ImportError("sisl cannot be imported. Please install sisl first.")
    fdf = sisl.get_sile(fdf_fname)
    H = fdf.read_hamiltonian()
    if H.spin.is_colinear:
        print("Reading Siesta hamiltonian: colinear spin.")
        tbmodel_up = SislWrapper(H, spin=0)
        tbmodel_dn = SislWrapper(H, spin=1)
        basis = dict(zip(tbmodel_up.orbs, list(range(tbmodel_up.norb))))
        print("Starting to calculate exchange.")
        exchange = ExchangeCL2(
            tbmodels=(tbmodel_up, tbmodel_dn),
            atoms=tbmodel_up.atoms,
            basis=basis,
            efermi=0.0,
            magnetic_elements=magnetic_elements,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            nz=nz,
            #height=height,
            #nz1=nz1,
            #nz2=nz2,
            #nz3=nz3,
            exclude_orbs=exclude_orbs,
            Rcut=Rcut,
            ne=ne,
            description=description)
        exchange.run()
        print("\n")
        print("All calculation finsihed. The results are in TB2J_results directory.")

    elif H.spin.is_colinear:
        print("Reading Siesta hamiltonian: colinear spin. Treat as non-colinear")
        tbmodel = SislWrapper(H, spin='merge')
        basis = dict(zip(tbmodel.orbs, list(range(tbmodel.nbasis))))
        print("Starting to calculate exchange.")
        exchange = ExchangeNCL(
            tbmodels=tbmodel,
            atoms=tbmodel.atoms,
            basis=basis,
            efermi=0.0,
            magnetic_elements=magnetic_elements,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            nz=nz,
            #height=height,
            #nz1=nz1,
            #nz2=nz2,
            #nz3=nz3,
            exclude_orbs=exclude_orbs,
            Rcut=Rcut,
            ne=ne,
            description=description)
        exchange.run()
        print("\n")
        print("All calculation finsihed. The results are in TB2J_results directory.")

    elif H.spin.is_spinorbit:
        print("Reading Siesta hamiltonian: non-colinear spin.")
        tbmodel = SislWrapper(H, spin=None)
        basis = dict(zip(tbmodel.orbs, list(range(tbmodel.nbasis))))
        print("Starting to calculate exchange.")
        exchange = ExchangeNCL(
            tbmodels=tbmodel,
            atoms=tbmodel.atoms,
            basis=basis,
            efermi=0.0,
            magnetic_elements=magnetic_elements,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            nz=nz,
            exclude_orbs=exclude_orbs,
            Rcut=Rcut,
            ne=ne,
            description=description)
        exchange.run()
        print("\n")
        print("All calculation finsihed. The results are in TB2J_results directory.")


def gen_exchange_gpaw(
                        gpw_fname,
                        magnetic_elements=[],
                        kmesh=[3, 3, 3],
                        emin=-12.0,
                        emax=0.0,
                        nz=50,
                        exclude_orbs=[],
                        Rcut=None,
                        description=''):
    print("Reading from GPAW data and calculate electronic structure.")
    model=GPAWWrapper(gpw_fname=gpw_fname)
    efermi=model.calc.get_fermi_level()
    print(f"Fermi Energy: {efermi}")
    poses=np.vstack([model.positions, model.positions])
    basis, _ = auto_assign_basis_name(poses, model.atoms)
    if model.calc.get_spin_polarized():
        print("Starting to calculate exchange.")
        exchange = ExchangeNCL(
            tbmodels=model,
            atoms=model.atoms,
            efermi=efermi,
            basis=basis,
            magnetic_elements=magnetic_elements,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            nz=nz,
            exclude_orbs=exclude_orbs,
            Rcut=Rcut,
            description=description)
        exchange.run()
        print("\n")
        print("All calculation finsihed. The results are in TB2J_results directory.")




"""
Exchange from Green's function
"""

from collections import defaultdict
import os
import numpy as np
from TB2J.green import TBGreen
from TB2J.utils import symbol_number, read_basis
from TB2J.myTB import MyTB
from ase.io import read
from TB2J.utils import auto_assign_basis_name
from TB2J.io_exchange import SpinIO
import progressbar
from functools import lru_cache
from .exchange import ExchangeCL


class ExchangeCL2(ExchangeCL):
    def set_tbmodels(self, tbmodels):
        """
        only difference is a colinear tag.
        """
        self.tbmodel_up, self.tbmodel_dn = tbmodels
        self.Gup = TBGreen(self.tbmodel_up, self.kmesh, self.efermi)
        self.Gdn = TBGreen(self.tbmodel_dn, self.kmesh, self.efermi)
        self.norb = self.Gup.norb
        self.nbasis = self.Gup.nbasis + self.Gdn.nbasis
        self.rho_up = np.zeros((self.norb, self.norb), dtype=float)
        self.rho_dn = np.zeros((self.norb, self.norb), dtype=float)
        self.JJ = defaultdict(lambda: 0.0j)
        self.Jorb = defaultdict(lambda: 0.0j)
        self.HR0_up = self.Gup.H0
        self.HR0_dn = self.Gdn.H0
        self.Delta = self.HR0_up - self.HR0_dn
        if self.Gup.is_orthogonal and self.Gdn.is_orthogonal:
            self.is_orthogonal =True
        else:
            self.is_orthogonal = False
            self.S0=self.Gup.S0
        self._is_colinear = True

        self.exchange_Jdict = {}
        self.exchange_Jdict_orb={}

        self.biquadratic = False

    def get_Delta(self, iatom):
        orbs = self.iorb(iatom)
        return self.Delta[np.ix_(orbs, orbs)]

    def GR_atom(self, GR, iatom, jatom):
        """Given a green's function matrix, return the [iatom, jatom] component.

        :param GR:  Green's function matrix
        :param iatom:  index of atom i
        :param jatom:  index of atom j
        :returns:  G_ij
        :rtype:  complex matrix.
        """
        orbi = self.iorb(iatom)
        orbj = self.iorb(jatom)
        return GR[np.ix_(orbi, orbj)]

    def get_A_ijR(self, Gup, Gdn, iatom, jatom, de):
        Rij_done=set()
        for R, ijpairs in self.R_ijatom_dict.items():
            if (iatom, jatom) in ijpairs and (R, iatom, jatom) not in Rij_done:
                Gij_up = self.GR_atom(Gup[R], iatom, jatom)
                Rm = tuple(-x for x in R)
                Gji_dn = self.GR_atom(Gdn[Rm], jatom, iatom)
                tmp = 0.0j
                #t = self.get_Delta(iatom) @ Gij_up @ self.get_Delta(jatom) @ Gji_dn
                t = np.einsum('ij, ji-> ij',
                              np.matmul(self.get_Delta(iatom), Gij_up),
                              np.matmul(self.get_Delta(jatom), Gji_dn))
                if self.biquadratic:
                    A = np.einsum('ij, ji-> ij',
                              np.matmul(self.get_Delta(iatom), Gij_up),
                              np.matmul(self.get_Delta(jatom), Gji_up))
                    C = np.einsum('ij, ji-> ij',
                              np.matmul(self.get_Delta(iatom), Gij_down),
                              np.matmul(self.get_Delta(jatom), Gji_down))
                tmp = np.sum(t)
                self.Jorb[(R, iatom, jatom)] += t * de / (4.0 * np.pi)
                self.JJ[(R, iatom, jatom)] += tmp * de / (4.0 * np.pi)
                Rij_done.add((R,iatom, jatom))
                if (Rm,jatom, iatom) not in Rij_done:
                    self.Jorb[(Rm, jatom, iatom)] += t * de / (4.0 * np.pi)
                    self.JJ[(Rm, jatom, iatom)] += tmp * de / (4.0 * np.pi)
                    Rij_done.add((Rm,jatom, iatom))

    def get_all_A(self, Gup, Gdn, de):
        """
        Calculate all A matrix elements
        Loop over all magnetic atoms.
        :param G: Green's function.
        :param de: energy step.
        """
        Rij_done=set()
        for R, ijpairs in self.R_ijatom_dict.items():
            for iatom, jatom in ijpairs:
                if (R, iatom, jatom) not in Rij_done:
                    Rm = tuple(-x for x in R)
                    if (Rm,jatom, iatom) in Rij_done:
                        raise KeyError(f"Strange (Rm, jatom, iatom) has already been calculated! {(Rm, jatom, iatom)}")
                    Gij_up = self.GR_atom(Gup[R], iatom, jatom)
                    Gji_dn = self.GR_atom(Gdn[Rm], jatom, iatom)
                    tmp = 0.0j
                    #t = self.get_Delta(iatom) @ Gij_up @ self.get_Delta(jatom) @ Gji_dn
                    t = np.einsum('ij, ji-> ij',
                                  np.matmul(self.get_Delta(iatom), Gij_up),
                                  np.matmul(self.get_Delta(jatom), Gji_dn))
                    tmp = np.sum(t)
                    self.Jorb[(R, iatom, jatom)] += t * de / (4.0 * np.pi)
                    self.JJ[(R, iatom, jatom)] += tmp * de / (4.0 * np.pi)
                    Rij_done.add((R,iatom, jatom))
                    if (Rm,jatom, iatom)  not in Rij_done:
                        self.Jorb[(Rm, jatom, iatom)] += t * de / (4.0 * np.pi)
                        self.JJ[(Rm, jatom, iatom)] += tmp * de / (4.0 * np.pi)
                        Rij_done.add((Rm,jatom, iatom))
                            

    def A_to_Jtensor(self):
        for key, val in self.JJ.items():
            # key:(R, iatom, jatom)
            R, iatom, jatom = key
            ispin = self.ispin(iatom)
            jspin = self.ispin(jatom)
            keyspin = (R, ispin, jspin)
            is_nonself = not (R == (0, 0, 0) and iatom == jatom)
            Jij = np.imag(val) / np.sign(
                np.dot(self.spinat[iatom], self.spinat[jatom]))
            Jorbij=np.imag(self.Jorb[key])/np.sign(
                np.dot(self.spinat[iatom], self.spinat[jatom]))
            if is_nonself:
                self.exchange_Jdict[keyspin] = Jij
                self.exchange_Jdict_orb[keyspin] =Jorbij

    def get_rho_e(self, rho_up, rho_dn, de):
        #GR0_up = GR_up[(0, 0, 0)]
        #GR0_dn = GR_dn[(0, 0, 0)]
        #if self.is_orthogonal:
        #    self.rho_up += -1.0 / np.pi * np.imag(GR0_up * de)
        #    self.rho_dn += -1.0 / np.pi * np.imag(GR0_dn * de)
        #else:
        #    self.rho_up += -1.0 / np.pi * np.imag(self.S0@GR0_up * de)
        #    self.rho_dn += -1.0 / np.pi * np.imag(self.S0@GR0_dn * de)
        self.rho_up += -1.0 / np.pi * np.imag(rho_up[(0,0,0)] * de)
        self.rho_dn += -1.0 / np.pi * np.imag(rho_dn[(0,0,0)] * de)

    def get_rho_atom(self):
        """
        charges and spins from density matrices
        """
        self.charges = np.zeros(len(self.atoms), dtype=float)
        self.spinat = np.zeros((len(self.atoms), 3), dtype=float)
        for iatom in self.orb_dict:
            iorb = self.iorb(iatom)
            tup = np.real(np.trace(self.rho_up[np.ix_(iorb, iorb)]))
            tdn = np.real(np.trace(self.rho_dn[np.ix_(iorb, iorb)]))
            # *2 because there is a 1/2 in the paui_block_all function
            self.charges[iatom] = tup + tdn
            self.spinat[iatom, 2] = tup - tdn

    def calculate_all(self):
        """
        The top level.
        """
        print("Green's function Calculation started.")

        widgets = [
            ' [',
            progressbar.Timer(),
            '] ',
            progressbar.Bar(),
            ' (',
            progressbar.ETA(),
            ') ',
        ]
        bar = progressbar.ProgressBar(
            maxval=self.contour.npoints, widgets=widgets)
        bar.start()
        for ie in range(self.contour.npoints):
            bar.update(ie)
            e = self.contour.path[ie]
            de = self.contour.de[ie]
            GR_up, rho_up = self.Gup.get_GR(self.short_Rlist, energy=e, get_rho=True)
            GR_dn, rho_dn = self.Gdn.get_GR(self.short_Rlist, energy=e, get_rho=True)
            self.get_rho_e(rho_up, rho_dn, de)
            self.get_all_A(GR_up, GR_dn, de)
        self.get_rho_atom()
        self.A_to_Jtensor()
        bar.finish()

    def write_output(self, path='TB2J_results'):
        self._prepare_index_spin()
        output = SpinIO(
            atoms=self.atoms,
            charges=self.charges,
            spinat=self.spinat,
            index_spin=self.index_spin,
            colinear=True,
            distance_dict=self.distance_dict,
            exchange_Jdict=self.exchange_Jdict,
            exchange_Jdict_orb = self.exchange_Jdict_orb,
            dmi_ddict=None,
            NJT_Jdict=None,
            NJT_ddict=None,
            Jani_dict=None,
            biquadratic_Jdict=None,
        )
        output.write_all()



import numpy as np
import scipy.linalg as sl
from collections import defaultdict
from ase.dft.kpoints import monkhorst_pack
import os


def eigen_to_G(evals, evecs, efermi, energy):
    """ calculate green's function from eigenvalue/eigenvector for energy(e-ef): G(e-ef).
    :param evals:  eigen values
    :param evecs:  eigen vectors
    :param efermi: fermi energy
    :param energy: energy
    :returns: Green's function G,
    :rtype:  Matrix with same shape of the Hamiltonian (and eigenvector)
    """
    return evecs.dot(np.diag(1.0 / (-evals + (energy + efermi)))).dot(
        evecs.conj().T)

class TBGreen():
    def __init__(
            self,
            tbmodel,
            kmesh,  # [ikpt, 3]
            efermi,  # efermi
            k_sym=False):
        """
        :param tbmodel: A tight binding model
        :param kmesh: size of monkhorst pack. e.g [6,6,6]
        :param efermi: fermi energy.
        """
        self.tbmodel = tbmodel
        self.R2kfactor=tbmodel.R2kfactor
        self.k2Rfactor=-tbmodel.R2kfactor
        self.efermi = efermi
        if kmesh is not None:
            self.kpts = monkhorst_pack(size=kmesh)
        else:
            self.kpts = tbmodel.get_kpts()
        self.nkpts = len(self.kpts)
        self.kweights = [1.0 / self.nkpts] * self.nkpts
        self.norb = tbmodel.norb
        self.nbasis = tbmodel.nbasis
        self.k_sym=k_sym
        self._prepare_eigen()

    def _prepare_eigen(self):
        """
        calculate eigen values and vectors for all kpts and save.
        Note that the convention 2 is used here, where the 
        phase factor is e^(ik.R), not e^(ik.(R+rj-ri))
        """
        # self.evals = np.zeros((len(self.kpts), self.nbasis), dtype=complex)
        # self.evecs = np.zeros((len(self.kpts), self.nbasis, self.nbasis),
        #                       dtype=complex)
        # self.H0 = 0.0
        # for ik, k in enumerate(self.kpts):
        #     Hk = self.tbmodel.gen_ham(tuple(k))
        #     self.evals[ik, :], self.evecs[ik, :, :] =self.tbmodel.solve(tuple(k))
        #     self.H0 += Hk / len(self.kpts)
        H,S,self.evals, self.evecs=self.tbmodel.HS_and_eigen(self.kpts)
        self.H0=np.sum(H, axis=0)/len(self.kpts)
        if S is not None:
            self.is_orthogonal=False
            self.S=S
            self.S0=np.sum(S, axis=0)/len(self.kpts)
        else:
            self.is_orthogonal=True

    def get_Gk(self, ik, energy):
        """ Green's function G(k) for one energy
        G(\epsilon)= (\epsilon I- H)^{-1}
        :param ik: indices for kpoint
        :returns: Gk
        :rtype:  a matrix of indices (nbasis, nbasis)
        """
        Gk = eigen_to_G(
            evals=self.evals[ik, :],
            evecs=self.evecs[ik, :, :],
            efermi=self.efermi,
            energy=energy)
        # A slower version. For test.
        #Gk = np.linalg.inv((energy+self.efermi)*self.S[ik,:,:] - self.H[ik,:,:])
        return Gk

    def get_GR(self, Rpts, energy, get_rho=False):
        """ calculate real space Green's function for one energy, all R points.
        G(R, epsilon) = G(k, epsilon) exp(-2\pi i R.dot. k)
        :param Rpts: R points
        :param energy:
        :returns:  real space green's function for one energy for a list of R.
        :rtype:  dictionary, the keys are tuple of R, values are matrices of nbasis*nbasis
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0j)
        rhoR = defaultdict(lambda: 0.0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            if get_rho:
                if self.is_orthogonal:
                    rhok = Gk
                else:
                    rhok=self.S[ik]@Gk
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                tmp=Gk * (phase * self.kweights[ik])
                GR[R] += tmp
                if get_rho:
                    rhoR[R]+=rhok*(phase * self.kweights[ik])
        if get_rho:
            return GR, rhoR
        else:
            return GR

    def get_GR_and_dGRdx1(self, Rpts, energy, dHdx):
        """
        calculate G(R) and dG(R)/dx.
        dG(R)/dx = \sum_k G(k) (dH(R)/dx) G(k).
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0 + 0.0j)
        dGRdx = defaultdict(lambda: 0.0 + 0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            Gkw = Gk * self.kweights[ik]
            #Gmk = self.get_Gk(self.i_minus_k(kpt), energy)
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                GR[R] += Gkw * (phase * self.kweights[ik])

                dHRdx = dHdx.get_hamR(R)
                dGRdx[R] += Gkw @ dHRdx @ Gk
                #dGRdx[R] += Gk.dot(dHRdx).dot(Gkp)
        return GR, dGRdx

    def get_GR_and_dGRdx(self, Rpts, energy, dHdx):
        """
        calculate G(R) and dG(R)/dx.
        dG(k)/dx =  G(k) (dH(k)/dx) G(k).
        dG(R)/dx = \sum_k dG(k)/dx * e^{-ikR}
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0 + 0.0j)
        dGRdx = defaultdict(lambda: 0.0 + 0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            #Gmk = self.get_Gk(self.i_minus_k(kpt), energy)
            Gkp = Gk * self.kweights[ik]
            dHk = dHdx.gen_ham(tuple(kpt))
            dG = Gk @ dHk @ Gkp
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                GR[R] += Gkp * (phase*self.kweights[ik])
                dGRdx[R] += dG * (phase*self.kweights[ik])
        return GR, dGRdx

    def get_GR_and_dGRdx_and_dGRdx2(self, Rpts, energy, dHdx, dHdx2):
        """
        calculate G(R) and dG(R)/dx.
        dG(k)/dx =  G(k) (dH(k)/dx) G(k).
        dG(R)/dx = \sum_k dG(k)/dx * e^{-ikR}
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0 + 0.0j)
        dGRdx = defaultdict(lambda: 0.0 + 0j)
        dGRdx2 = defaultdict(lambda: 0.0 + 0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            #Gmk = self.get_Gk(self.i_minus_k(kpt), energy)
            Gkp = Gk * self.kweights[ik]
            dHk = dHdx.gen_ham(tuple(kpt))
            dHk2 = dHdx2.gen_ham(tuple(kpt))
            dG = Gk @ dHk @ Gkp
            dG2 = Gk @ dHk2 @ Gkp
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                GR[R] += Gkp * (phase *self.kweights[ik])
                dGRdx[R] += dG * (phase*self.kweights[ik])
                dGRdx2[R] += dG2 * (phase*self.kweights[ik])
        return GR, dGRdx, dGRdx2





import os
import pickle
import copy
import numpy as np
from scipy.spatial.transform import Rotation
from TB2J.io_exchange import SpinIO

# Rotation from x to z
Rxz = Rotation.from_euler('y', -90, degrees=True)
# Rotation from y to z
Ryz = Rotation.from_euler('x', 90, degrees=True)

def merge_DMI(Dx, Dy, Dz):
    Dx_z = Rxz.apply(Dx)
    Dy_z = Ryz.apply(Dy)
    D = (np.array([0.0, 0.5, 0.5]) * Dx_z + np.array([0.5, 0.0, 0.5]) * Dy_z +
         np.array([0.5, 0.5, 0.0]) * Dz)
    return D


class Merger():
    def __init__(self, path_x, path_y, path_z):
        self.dat_x=SpinIO.load_pickle(os.path.join(path_x, 'TB2J_results'), 'TB2J.pickle')
        self.dat_y=SpinIO.load_pickle(os.path.join(path_y, 'TB2J_results'), 'TB2J.pickle')
        self.dat_z=SpinIO.load_pickle(os.path.join(path_z, 'TB2J_results'), 'TB2J.pickle')
        self.dat=copy.copy(self.dat_z)

    def merge_DMI(self):
        dmi_ddict={}
        if self.dat_x.has_dmi and self.dat_y.has_dmi and self.dat_z.has_dmi:
            Dxdict=self.dat_x.dmi_ddict
            Dydict=self.dat_y.dmi_ddict
            Dzdict=self.dat_z.dmi_ddict
            for key, Dz in Dzdict.items():
                try:
                    Dx= Dxdict[key]
                    Dy= Dydict[key]
                except KeyError as err:
                    raise KeyError("%s, Please make sure the three calculations use the same k-mesh and same Rcut."%err)
                dmi_ddict[key] = merge_DMI(Dx, Dy, Dz)
            self.dat.dmi_ddict=dmi_ddict

    def write(self, path='TB2J_results'):
        self.dat.write_all(path=path)



def merge(path_x, path_y, path_z, save=True, path='TB2J_results'):
    m=Merger(path_x, path_y, path_z)
    m.merge_DMI()
    if save:
        m.write(path=path)
    return m.dat




import matplotlib.pyplot as plt
from TB2J.spinham.spin_api import SpinModel


def plot_magnon_band(fname='exchange.xml',
                     npoints=301,
                     show=True,
                     kvectors=None,
                     knames=None,
                     figfname="magnon_band.pdf",
                     Jq=False,
                     kpath_fname='exchange_kpth.txt'):
    m = SpinModel(fname=fname, sc_matrix=None)
    m.plot_magnon_band(kvectors=kvectors,
                       knames=knames,
                       npoints=npoints,
                       kpath_fname=kpath_fname,
                       Jq=Jq,
)
    plt.savefig(figfname)
    if show:
        plt.show()



"""
The script to plot the magnon band structure.
"""

import argparse
from TB2J.versioninfo import print_license
from TB2J.plot import plot_magnon_band


def plot_magnon():
    print_license()
    parser = argparse.ArgumentParser(
        description=
        "TB2J_magnon: Plot magnon band structure from the TB2J magnetic interaction parameters"
    )
    parser.add_argument("--fname",
                        default='exchange.xml',
                        type=str,
                        help='exchange xml file name. default: exchange.xml')

    parser.add_argument(
        "--qpath",
        default=None,
        type=str,
        help=
        'The names of special q-points. If not given, the path will be automatically choosen. See https://wiki.fysik.dtu.dk/ase/ase/dft/kpoints.html for the table of special kpoints and the default path.'
    )


    parser.add_argument(
        "--figfname",
        default=None,
        type=str,
        help=
        'The file name of the figure. It should be e.g. png, pdf or other types of files which could be generated by matplotlib.'
    )
 
    parser.add_argument("--Jq",
                        action="store_true",
                        help="To plot the eigenvalues of -J(q) instead of the magnon band",
                        default=False)
 
    parser.add_argument("--show",
                        action="store_true",
                        help="whether to show magnon band structure.",
                        default=False)

    args = parser.parse_args()
    if args.Jq:
        print("Plotting the eigenvalues of -J(q). The figure is written to %s"%(args.figfname) )
        if args.figfname is None:
            args.figfname = 'Eigen_Jq.pdf'
    else:
        print("Plotting the magnon band structure. The figure is written to %s"%(args.figfname))
        if args.figfname is None:
            args.figfname = 'magnon_band.pdf'

    plot_magnon_band(fname=args.fname,
                     knames=args.qpath,
                     npoints=300,
                     Jq=args.Jq,
                     show=args.show)


plot_magnon()



import numpy as np
from ase.io import read
from ase.atoms import Atoms
from collections import defaultdict
import math

def parse_xyz(fname):
    atoms=read(fname)
    symbols=atoms.get_chemical_symbols()
    pos=atoms.get_positions()
    wann_pos=[]
    atoms_pos=[]
    atoms_symbols=[]
    for s, x in zip(symbols, pos):
        if s=='X':
            wann_pos.append(x)
        else:
            atoms_symbols.append(s)
            atoms_pos.append(x)
    return np.array(wann_pos), atoms_symbols, np.array(atoms_pos)

def parse_ham(fname='wannier90_hr.dat',cutoff=None):
    """
    wannier90 hr file phaser.

    :param cutoff: the energy cutoff.  None | number | list (of Emin, Emax).
    """
    with open(fname,'r') as myfile:
        lines=myfile.readlines()
    n_wann=int(lines[1].strip())
    n_R=int(lines[2].strip())

    # The lines of degeneracy of each R point. 15 per line.
    nline=int(math.ceil(n_R/15.0))
    dlist=[]
    for i in range(3,3+nline):
        d=map(float,lines[i].strip().split())
        dlist+=d
    H_mnR=defaultdict(lambda: np.zeros((n_wann, n_wann), dtype=complex))
    for i in range(3+nline,3+nline+n_wann**2*n_R):
        t=lines[i].strip().split()
        R=tuple(map(int,t[:3]))
        m,n=map(int,t[3:5])
        m=m-1
        n=n-1
        H_real,H_imag=map(float, t[5:])
        val=H_real+1j*H_imag
        if (m==n and np.linalg.norm(R)<0.001):
            H_mnR[R][m, n]=val/2.0
        elif cutoff is not None:
            if abs(val)>cutoff:
                H_mnR[R][m, n]=val
        else:
           H_mnR[R][m, n]=val/2.0
    return n_wann, H_mnR

